#include <Wire.h>
#include <ESP8266WiFi.h>
#include <PubSubClient.h> // Allows us to connect to, and publish to the MQTT broker
#include <ESP_EEPROM.h>


// Temperature Sensor Addresses
#define add1 0x48
#define add2 0x49

// Temperature Sensor Configuration Bits
#define highresOS 0b10001101
#define lowresOS 0b10000001
#define highresC 0b10001100
#define lowresC 0b10000000

// EEPROM Variable Setup
bool warnStat; // 8 bits
bool batStat; // 8 bits
bool sensorStat; // 8 bits
word dayCount; // 16 bits

int warnAdd = 0;
int batAdd = 1;
int sensorAdd = 2;
int countAdd = 3;
int error = -59;

// Test Variables
int sensorRange = 5; // Sensor Maintenence
int u_thresh = 10; // Upper Temperature Threshold
int l_thresh = 5; // Lower Temperature Threshold
int countmax = 1; // Deep sleep counts 

bool zero = 0;
float temp2;

// WiFi
// Make sure to update this for your own WiFi network!
const char* ssid = "ND-guest";

// MQTT
// Make sure to update this for your own MQTT Broker!
const char* mqtt_server = "73.50.54.233";
const char* mqtt_topic = "final";
const char* mqtt_username = "pipefreeze";
const char* mqtt_password = "almamater";
// The client id identifies the ESP8266 device. Think of it a bit like a hostname (Or just a name, like Greg).
const char* clientID = "esp8266room205";
const char* message; 

// Initialise the WiFi and MQTT Client objects
WiFiClient wifiClient;
PubSubClient client(mqtt_server, 1883, wifiClient); // 1883 is the listener port for the Broker

// Call MAC Address
String MAC = WiFi.macAddress();
String temp_s;
String code; 
String message_s;

void setup() {
  // put your setup code here, to run once:
  delay(100);
  Serial.begin(9600);

  Wire.begin(2,14);

  EEPROM.begin(16);
  //Serial.print("Percent EEPROM used: ");
  //Serial.println(EEPROM.percentUsed());
/*
  dayCount = 0;
  EEPROM.put(countAdd,dayCount);
  bool checkstuff = EEPROM.commit();
  Serial.println((checkstuff) ? "Check Commit ok" : "Check Commit failed");
*/
  EEPROM.get(countAdd,dayCount);
  
  Serial.println("Checking Day Count");
  
  if (dayCount <= countmax) {
    dayCount++;
    Serial.print("Keep Alive Counter: ");
    Serial.println(dayCount);
    EEPROM.put(countAdd,dayCount);
    bool okday = EEPROM.commit();
    Serial.println((okday) ? "Day1 Commit ok" : "Day1 Commit failed");
  }
  else {
    dayCount = 0;
    keepAlive();
    EEPROM.put(countAdd,dayCount);
    bool oknight = EEPROM.commit();
    Serial.println((oknight) ? "Day2 Commit ok" : "Day2 Commit failed");
  }

}


void EEPROMput(int add, bool stat) {
  EEPROM.put(add, stat);
  bool ok1 = EEPROM.commit();
  Serial.println((ok1) ? "Commit ok" : "Commit failed");
}

bool EEPROMget(int add) {
  bool temp;
  EEPROM.get(add, temp);
  return temp;
}

void keepAlive() {   
      temp_s = "NaN";
      code = "_0_";
      message_s = MAC+code+temp_s;

      wifi(message_s);
}

void wifi(String message_s) {
  WiFi.begin(ssid);
    
      // Wait until the connection has been confirmed before continuing
      while (WiFi.status() != WL_CONNECTED) {
        delay(10);
        Serial.print(".");
      }
    
      // Debugging - Output the IP Address of the ESP8266
      Serial.println("WiFi connected");
      Serial.print("IP address: ");
      Serial.println(WiFi.localIP());
    
      // Connect to MQTT Broker
      // client.connect returns a boolean value to let us know if the connection was successful.
      // If the connection is failing, make sure you are using the correct MQTT Username and Password (Setup Earlier in the Instructable)
      if (client.connect(clientID, mqtt_username, mqtt_password)) {
        Serial.println("Connected to MQTT Broker!");
      }
      else {
        Serial.println("Connection to MQTT Broker failed...");
      }
    
      /*temp_s = "NaN_";
      code = "_0_";
      
      message_s = MAC+code+temp_s;*/
      message = message_s.c_str();
    
      bool succeedornot = client.publish(mqtt_topic, message);
}


void loop() {
  // put your main code here, to run repeatedly:

  
  float u_temp;
  float l_temp;
  float temp1_sum = 0;
  int temp1_count = 0;
  float temp2_sum = 0;
  int temp2_count = 0;

  float temp1_array[5];
  float temp2_array[5];
  float temp1_max;
  float temp1_min;
  float temp2_max;
  float temp2_min;

  // Set battery monitor input pin
  int monPin = 4;
  int mon;
  pinMode(monPin, INPUT);
  
  // one shot
  Wire.beginTransmission(add1);
  Wire.write(0xAC);
  Wire.write(highresOS);
  Wire.endTransmission();

  delay(10);

  Wire.beginTransmission(add2);
  Wire.write(0xAC);
  Wire.write(highresOS);
  Wire.endTransmission();


  for (int i=0; i<5; i++) {

  // take temps
  Wire.beginTransmission(add1);
  Wire.write(0x51); // start temp conversion command
  Wire.endTransmission();

  delay(10);

  Wire.beginTransmission(add2);
  Wire.write(0x51); // start temp conversion command
  Wire.endTransmission();

  delay(10);
  
  Wire.beginTransmission(add1);
  Wire.write(0xAA); // read temp command
  Wire.endTransmission();
  Wire.requestFrom(add1, 2);
  Wire.available();
  uint8_t t1 = Wire.read();
  Wire.available();
  uint8_t t2 = Wire.read();
  Wire.endTransmission();
  
  // convert to celsius
  int16_t temp1_ = (t1 << 8) + t2;
  if (temp1_ >= 0x8000) { // check if temperature is negative
    temp1_ = temp1_ - 0xFFFF;
  }
  temp1_ >>= 4;
  float temp1 = temp1_ * 0.0625;

  Serial.print("Sensor 1 = ");
  Serial.println(temp1);

  delay(10);

  Wire.beginTransmission(add2);
  Wire.write(0xAA); // read temp command
  Wire.endTransmission();
  Wire.requestFrom(add2, 2);
  Wire.available();
  uint8_t t3 = Wire.read();
  Wire.available();
  uint8_t t4 = Wire.read();
  Wire.endTransmission();
  
  int16_t temp2_ = (t3 << 8) + t4;
  if (temp2_ >= 0x8000) { // check if temperature is negative
    temp2_ = temp2_ - 0xFFFF;
  }
  temp2_ >>= 4;
  float temp2 = temp2_ * 0.0625;
  
  
  Serial.print("Sensor 2 = ");
  Serial.println(temp2);

  delay(10);

  temp1_array[i] = temp1;
  temp2_array[i] = temp2;

  }

  temp1_max = temp1_array[0];
  temp1_min = temp1_array[0];
  temp2_max = temp2_array[0];
  temp2_min = temp2_array[0];
  
  for (int j=0; j<4; j++) {
    if (temp1_array[j] > temp1_max) {
      temp1_max = temp1_array[j];
    }

    if (temp1_array[j] < temp1_min) {
      temp1_min = temp1_array[j];
    }

    if (temp2_array[j] > temp2_max) {
      temp2_max = temp2_array[j];
    }

    if (temp2_array[j] < temp2_min) {
      temp2_min = temp2_array[j];
    }

    temp1_sum = temp1_array[j] + temp1_sum;
    temp2_sum = temp2_array[j] + temp2_sum;
    temp1_count = temp1_count + 1;
    temp2_count = temp2_count + 1;
  }
  
  float temp1_avg = (temp1_sum - temp1_max - temp1_min)/(temp1_count-2);
  float temp2_avg = (temp2_sum - temp2_max - temp2_min)/(temp2_count-2);
  float temp_avg = (temp1_avg+temp2_avg)/2;

  // sensor reliability check variables
  u_temp = temp1_avg + sensorRange;
  l_temp = temp1_avg - sensorRange;
  
  if ((temp2_avg > u_temp || temp2_avg < l_temp)) { // Sensor Maintenance Loop
    Serial.println("Entering sensor maintenance loop");

    sensorStat = EEPROMget(sensorAdd);

    if (!sensorStat) {
    
      temp_s = "NaN";
      code = "_-2_";
      //message_s = MAC+code+temp_s;
      String temp2str = String(temp2_avg,1);
      String temp1str = String(temp1_avg,1);
      message_s = MAC+code+temp_s+"_"+temp1str+"_"+temp2str;
      wifi(message_s);

      EEPROMput(sensorAdd,1);

    }
  
    Serial.println("Entering deep Sleep");
    ESP.deepSleep(12e6);
  
  }
  

  if (temp_avg < l_thresh && temp_avg > error) { // Critical Alert Loop
    Serial.println("Entering critical alert loop");
  
    temp_s = String(temp_avg,1); 
    code = "_2_";
    
    //message_s = MAC+code+temp_s;
    String temp2str = String(temp2_avg,1);
    String temp1str = String(temp1_avg,1);
    message_s = MAC+code+temp_s+"_"+temp1str+"_"+temp2str;
    wifi(message_s);
  
    Serial.println("Entering deep Sleep");
    ESP.deepSleep(12e6);
  }
  else if (temp_avg < u_thresh && temp_avg > error) { // Non-critical Alert Loop
    Serial.println("Entering non-critical alert loop");

    warnStat = EEPROMget(warnAdd);

    if (!warnStat) {
    
      temp_s = String(temp_avg,1); 
      code = "_1_";
      //message_s = MAC+code+temp_s;
      String temp2str = String(temp2_avg,1);
      String temp1str = String(temp1_avg,1);
      message_s = MAC+code+temp_s+"_"+temp1str+"_"+temp2str;
      wifi(message_s);

      EEPROMput(warnAdd,1);
    }
  
    Serial.println("Entering deep Sleep");
    ESP.deepSleep(12e6);
  }

  
  mon = digitalRead(monPin);

  if (mon == LOW && temp_avg > 14) { // Low Battery Alert Loop

    Serial.println("Entering Low Battery Alert Loop");

    batStat = EEPROMget(batAdd);

    if (!batStat) {
    
      temp_s = "NaN"; 
      code = "_-1_";
      //message_s = MAC+code+temp_s;
      String temp2str = String(temp2_avg,1);
      String temp1str = String(temp1_avg,1);
      message_s = MAC+code+temp_s+"_"+temp1str+"_"+temp2str;
      wifi(message_s);

      EEPROMput(batAdd,1);
    }
  
    Serial.println("Entering deep Sleep");
    ESP.deepSleep(12e6);
  
  }

  EEPROM.get(countAdd,dayCount);
  Serial.println(dayCount);
  delay(10);

  for (int i = 0; i < 3; i++) {
    EEPROM.put(i,zero);
  }
  bool ok2 = EEPROM.commit();
  Serial.println((ok2) ? "Reset ok" : "Reset failed");

  delay(10);
  EEPROM.get(countAdd,dayCount);
  Serial.println(dayCount);
  

  Serial.println("Entering deep sleep");
  ESP.deepSleep(12e6);

 
}
